"""
Implements the functionality of the context manager
used for code generation.
Naming might be suboptimal, but, did I say this is a POC?
"""


import sys
import inspect
import ast
import copy
from typing import List, Dict, Union
from cogpyt.exception import SkipWithBlockException


class GeneratedCodeBlock:
    """
    TODO: interface documentation

    TODO: pass statements not removed, as not harmful.
    Feel free to contribute to resolve if it bothers you.
    """

    def __init__(
            self,
            indent: int = 4,
            outer_indentation: bool = False,
            banned_locals: List[str] = None
    ):
        """
        Parameters
        ----------
        indent
            amount of spaces to use for code indentation
        outer_indentation
            whether to indent the entire exported code.
            Useful to just append the generated code
            to a function definition header.
        banned_locals
            Local variables to never replace in generated code.
            Currently only used to prevent replacing method arguments in code.
        """
        # NOTE: indentation starts at 1 to export source which can be easily
        #       appended to a function definition. After all, this is POC
        self._indentation = 0
        self._spaces_per_indentation = indent
        self._outer_indentation = outer_indentation
        self._banned_locals = banned_locals or []
        self._generated_code = []

        self._old_trace = None
        self._filename = None
        self._source_ast_nodes_by_line = None
        self._current_block_start = None
        self._current_locals = None

    def __enter__(self):
        """
        Sets an exception to be executed next in the `with` block.
        This allows the code in the with block not to be executed,
        but instead to be extracted and enriched as generated code
        in the `__exit__` method. Also keeps track of the relevant
        context in the moment of execution to do so.
        """
        # prevent with block execution
        self._old_trace = sys.gettrace()
        sys.settrace(lambda *args, **keys: None)
        context_entering_frame = inspect.currentframe().f_back
        context_entering_frame.f_trace = self._trace

        # store information on where the source of the with block starts
        # and which local variables are available at the time of execution
        self._current_block_start = context_entering_frame.f_lineno
        self._current_locals = context_entering_frame.f_locals

        filename = context_entering_frame.f_code.co_filename
        if not self._filename:
            self._filename = filename
            self._source_ast_nodes_by_line = self._get_source_ast_node_by_line(
                filename
            )
        else:
            assert self._filename == filename, (
                'Expected a single code generation context '
                'instance to be used only '
                'within the same source code file.'
            )

    @classmethod
    def _get_source_ast_node_by_line(
            cls,
            filename: str
    ) -> Dict[int, ast.With]:
        """
        Returns a dictionary of (line number -> with statement AST).
        """
        ast_node_by_line = {}

        with open(filename, 'r', encoding='utf8') as file:
            source = file.read()

        source_ast = ast.parse(source)
        for node in ast.walk(source_ast):
            if isinstance(node, ast.With):
                line = node.lineno
                # We assume there can't be multiple `with`
                # statements defined on the same line
                assert line not in ast_node_by_line, (
                    'Multiple with statements on the same source line.'
                )
                ast_node_by_line[line] = node

        return ast_node_by_line

    def _trace(self, frame, event, arg):
        sys.settrace(self._old_trace)
        raise SkipWithBlockException()

    def __exit__(self, type_, value, traceback) -> bool:
        assert issubclass(type_, SkipWithBlockException)

        code_lines = self._extract_generated_code()
        self._generated_code.extend(code_lines)

        return True

    def _extract_generated_code(self) -> List[str]:
        """
        Extracts the generated code from the `with` statement
        marked by the `self._current_block_start` line in the
        abstract syntax tree of the source code file,
        available from `self._source_ast_nodes_by_line`.
        Also replaces any constants available as local constants
        from `self._current_locals`.

        Returns
        -------
        The so generated code.
        """
        with_node = self._source_ast_nodes_by_line[self._current_block_start]
        # The same line of the source can result in multiple lines
        # in the generated code. Handle each one separately (i.e. copy).
        with_node_copy = copy.deepcopy(with_node)
        code_ast = ast.Module(with_node_copy.body, [])

        inline_locals_transform = _InlineLocalsNodeTransformer(
            self._current_locals,
            self._banned_locals,
        )
        inline_locals_transform.visit(code_ast)

        code_lines = ast.unparse(code_ast).split('\n')
        return self._indent_lines(code_lines, self._indentation)

    def _indent_lines(self, lines: List[str], indentation: int) -> List[str]:
        indentation_string = ' ' * indentation * self._spaces_per_indentation
        return [
            indentation_string + line
            for line in lines
        ]

    @classmethod
    def _transform_into_generated_code(
            cls,
            source: str,
            indentation: int
    ) -> List[str]:
        pass

    def indent(self, amount: int = 1):
        """
        Increments the indentation level of the following
        generated code.
        """
        self._verify_indentation_amount(amount)
        self._indentation += amount

    def dedent(self, amount: int = 1):
        """
        Decrements the indentation level of the following
        generated code.
        """
        self._verify_indentation_amount(amount)

        if self._indentation < amount:
            raise ValueError('Cannot dedent past minimum indentation!')

        self._indentation -= amount

    @classmethod
    def _verify_indentation_amount(cls, amount):
        assert amount > 0, 'Indentation amount needs to be positive'
        assert type(amount) == int, 'Indentation amount needs to be an integer'

    def export(self) -> str:
        """
        Returns
        -------
        The source code generated thus far.
        """
        code = self._generated_code
        if self._outer_indentation:
            code = self._indent_lines(code, 1)

        return '\n'.join(code)


class _InlineLocalsNodeTransformer(ast.NodeTransformer):
    """
    Replaces all variables (ast.Name), which have defined substitutable values
    in the local variables scope, with constants (ast.Constant).
    This allows constants to be hard-coded in the generated code.
    A value is considered substitutable if it can be represented as
    a literal in pure python code.
    """
    def __init__(self, locals_: dict, banned_locals: List[str]):
        self.locals = {
            key: value
            for key, value in locals_.items()
            if key not in banned_locals and self.is_substitutable(value)
        }

    @classmethod
    def is_substitutable(cls, value) -> bool:
        """
        Returns
        -------
        Whether the value is substitutable directly in source code,
        i.e. whether it can be inlined.
        """
        try:
            # TODO: any better approach for this?
            # pylint: disable=eval-used
            eval(repr(value))
            return True
        except SyntaxError:
            return False

    # pylint: disable=invalid-name
    def visit_Name(self, name: ast.Name) -> Union[ast.Name, ast.Constant]:
        """
        Visitor method, this is the main workhorse.
        Replaces variables (names) with constants where possible.
        """
        if name.id in self.locals:
            value = self.locals[name.id]
            return ast.Constant(value)

        return name



import pytest
from cogpyt.context import GeneratedCodeBlock


@pytest.fixture
def generated_code_block():
    return GeneratedCodeBlock()


def test_generated_code_not_executed_on_generation(generated_code_block):
    prior_code_executed = True

    code_in_generated_block_executed = False
    with generated_code_block:
        code_in_generated_block_executed = True

    post_code_executed = True

    assert prior_code_executed
    assert not code_in_generated_block_executed
    assert post_code_executed


def test_generated_code_extraction(generated_code_block):
    with generated_code_block:
        print('some code')

    result = generated_code_block.export()
    expected = "print('some code')"

    assert result == expected


def test_generation_replaces_string_constants(generated_code_block):
    string = 'string'
    with generated_code_block:
        print(string)

    result = generated_code_block.export()
    expected = "print('string')"

    assert result == expected


def test_generation_replaces_numeric_constants(generated_code_block):
    number = 42.1337
    with generated_code_block:
        print(number)

    result = generated_code_block.export()
    expected = "print(42.1337)"

    assert result == expected


def test_generation_replaces_list_constants(generated_code_block):
    lst = [1, 3, 3, 7]
    with generated_code_block:
        print(lst)

    result = generated_code_block.export()
    expected = "print([1, 3, 3, 7])"

    assert result == expected


def test_generation_does_not_replace_instance_constants(generated_code_block):
    instance = object()
    with generated_code_block:
        print(instance)

    result = generated_code_block.export()
    expected = "print(instance)"

    assert result == expected


def test_generation_allows_multiple_blocks(generated_code_block):
    with generated_code_block:
        print('first line')

    with generated_code_block:
        print('second line')

    result = generated_code_block.export()
    expected = '\n'.join([
        "print('first line')",
        "print('second line')",
    ])

    assert result == expected


def test_generation_respects_indentation(generated_code_block):
    generated_code_block.indent()
    with generated_code_block:
        print('code')

    result = generated_code_block.export()
    expected = "    print('code')"

    assert result == expected


def test_generation_keeps_indentation(generated_code_block):
    generated_code_block.indent()
    with generated_code_block:
        print('first line')

    with generated_code_block:
        print('second line')

    result = generated_code_block.export()
    expected = '\n'.join([
        "    print('first line')",
        "    print('second line')",
    ])

    assert result == expected


def test_generation_allows_dedentation(generated_code_block):
    generated_code_block.indent()
    with generated_code_block:
        print('first line')

    generated_code_block.dedent()
    with generated_code_block:
        print('second line')

    result = generated_code_block.export()
    expected = '\n'.join([
        "    print('first line')",
        "print('second line')",
    ])

    assert result == expected


def test_generation_ignores_passes(generated_code_block):
    with generated_code_block:
        for i in range(42):
            pass

    generated_code_block.indent()
    with generated_code_block:
        print(i)

    result = generated_code_block.export()
    expected = '\n'.join([
        "for i in range(42):",
        "    pass",
        "    print(i)",
    ])

    assert result == expected


def test_mixing_generator_and_generated_variables_is_a_bad_idea(
        generated_code_block
):
    var = 42
    with generated_code_block:
        var += 1
        return var

    result = generated_code_block.export()
    expected = '\n'.join([
        '42 += 1',
        'return 42'
    ])

    assert result == expected



"""
Cogpyt implements code generation in pure Python.
"""


from .__about__ import __title__, __author__, __description__
from .__about__ import __url__, __version__, __copyright__, __licence__
from .decorator import GeneratedFunction
from .context import GeneratedCodeBlock


__all__ = [
    '__title__',
    '__author__',
    '__description__',
    '__url__',
    '__version__',
    '__copyright__',
    '__licence__',
    'GeneratedFunction',
    'GeneratedCodeBlock',
]



#!/usr/bin/env python


import pathlib
import setuptools


requires = [
    # nothing thus far ;)
]

root = pathlib.Path(__file__).parent.resolve()
with (root / 'README.md').open('r', encoding='utf-8') as f:
    readme = f.read()

about = {}
with (root / 'cogpyt' / '__about__.py').open('r', encoding='utf-8') as f:
    exec(f.read(), about)

setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    author=about['__author__'],
    long_description=readme,
    license=about['__licence__'],
    long_description_content_type='text/markdown',
    url=about['__url__'],
    packages=setuptools.find_packages(),
    package_data={'': ['LICENSE']},
    package_dir={'cogpyt': 'cogpyt'},
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=requires,
)



"""
Holds general information about the project.
"""

__title__ = 'cogpyt'
__author__ = 'Boyan Hristov'
__description__ = "Code Generation in (pure) Python"
__url__ = 'https://github.com/BoyanH/cogpyt'
__version__ = '0.0.1'
__copyright__ = 'Copyright 2022 Boyan Hristov'
__licence__ = 'MIT'



from typing import List
from itertools import count
import pytest
from cogpyt.decorator import GeneratedFunction
from cogpyt.context import GeneratedCodeBlock


# Not the cleanest, the wrapper is adding this argument.
# You should have come across the term POC in the project already ;)
# pylint: disable=no-value-for-parameter


def test_source_investigation():
    @GeneratedFunction
    def printer_method(
            generated_code_block: GeneratedCodeBlock,
            *print_texts: List[str],
    ):
        for text in print_texts:
            with generated_code_block:
                print(text)

    generated_code = printer_method.get_source(
        'Hello', 'code', 'generation', 'in', 'pure', 'Python', '!'
    )
    expected = '\n'.join([
        "def printer_method(*print_texts):",
        "    print('Hello')",
        "    print('code')",
        "    print('generation')",
        "    print('in')",
        "    print('pure')",
        "    print('Python')",
        "    print('!')",
    ])

    assert generated_code == expected


def test_generated_returns():
    @GeneratedFunction
    def generated_identity(
            generated_code_block: GeneratedCodeBlock,
            argument: any,
    ):
        with generated_code_block:
            return argument

    returned_value = generated_identity(42)
    assert returned_value == 42


def test_generated_modifies_arguments():
    @GeneratedFunction
    def generated_elementwise_incrementer(
            generated_code_block: GeneratedCodeBlock,
            lst: List[int],
    ):
        with generated_code_block:
            for i, element in enumerate(lst):
                lst[i] = element + 1

    lst = [4, 2]
    generated_elementwise_incrementer(lst)
    print(generated_elementwise_incrementer.get_source(lst))
    expected = [5, 3]
    assert lst == expected


def test_generated_propagates_exceptions():
    @GeneratedFunction
    def generated_raiser(
            generated_code_block: GeneratedCodeBlock,
    ):
        with generated_code_block:
            raise Exception

    with pytest.raises(Exception):
        generated_raiser()


def test_generated_is_not_closure():
    """
    Generated methods can only inline simple
    variables from the generator code,
    can't access or modify others.
    Cogpy won't help you implement fancier closures!
    """

    @GeneratedFunction
    def generated_not_closure_counter(
            generated_code_block: GeneratedCodeBlock,
    ):
        counter = count()
        with generated_code_block:
            return next(counter)

    with pytest.raises(NameError):
        generated_not_closure_counter()


def test_generated_passes_are_irrelevant():
    @GeneratedFunction
    def generated_counter_with_passes(
            generated_code_block: GeneratedCodeBlock,
    ):
        with generated_code_block:
            counter = 0
            for _ in range(10):
                counter += 1

            return counter

    assert generated_counter_with_passes() == 10


def test_mixing_generator_and_generated_variables_is_a_bad_idea():
    @GeneratedFunction
    def generated_with_mixed_variables(
            generated_code_block: GeneratedCodeBlock,
    ):
        var = 42
        with generated_code_block:
            var += 1
            return var

    with pytest.raises(SyntaxError):
        generated_with_mixed_variables()



"""
Contains the functionality for the library's decorator.
Naming might be suboptimal, but, did I say this is a POC?
"""


import inspect
from typing import Callable
from .context import GeneratedCodeBlock


class GeneratedFunction:
    """
    Represents the generated function.
    You can call it directly, or use `get_source()`
    to investigate the generated source and debug it.
    """
    def __init__(self, func: Callable):
        self._generator_function = func
        self._signature = self._get_signature(func)
        self._name = func.__name__

    @property
    def name(self):
        return self._name

    @property
    def generator_function(self):
        return self._generator_function

    @classmethod
    def _get_signature(cls, func):
        # includes the generated_code_block parameter for code generation
        # as can be seen in `get_source()`
        modified_signature = inspect.signature(func)

        modified_parameters = modified_signature.parameters
        tail_parameters = list(modified_parameters)[1:]
        # also, remove any signatures, since these aren't really handled
        # perfectly in python yet...
        original_parameters = [
            modified_parameters[p].replace(annotation=inspect.Signature.empty)
            for p in tail_parameters
        ]

        return modified_signature.replace(
            parameters=original_parameters,
            return_annotation=inspect.Signature.empty,
        )

    def __call__(self, *args, **kwargs) -> any:
        scope = {}
        source = self.get_source(*args, **kwargs)
        # sure, project is a HACK.
        # Wouldn't be writing this at 3AM if it wasn't.
        # pylint: disable=exec-used
        exec(source, scope)
        return scope[self._name](*args, **kwargs)

    def get_source(self, *args, **kwargs) -> str:
        """
        Returns
        -------
        The generated Python source code for
        the given set of arguments.
        """
        argument_names = list(self._signature.parameters.keys())
        generated_code_block = GeneratedCodeBlock(
            outer_indentation=True,
            banned_locals=argument_names
        )
        self.generator_function(generated_code_block, *args, **kwargs)

        return '\n'.join([
            f'def {self._name}{self._signature}:',
            generated_code_block.export()
        ])
